from pymobiledevice3.services.afc import AfcService

TEST_FILENAME = 'test'


def test_file_read_write(lockdown):
    afc = AfcService(lockdown)
    body = b'data'

    afc.set_file_contents(TEST_FILENAME, body)
    assert afc.get_file_contents(TEST_FILENAME) == body


def test_ls(lockdown):
    afc = AfcService(lockdown)
    filenames = afc.listdir('/')
    assert 'DCIM' in filenames


def test_rm(lockdown):
    afc = AfcService(lockdown)
    afc.set_file_contents(TEST_FILENAME, b'')
    filenames = afc.listdir('/')
    assert TEST_FILENAME in filenames

    afc.rm(TEST_FILENAME)

    filenames = afc.listdir('/')
    assert TEST_FILENAME not in filenames



# flake8: noqa: C901
import os

import click
from termcolor import colored

from pymobiledevice3.cli.cli_common import Command
from pymobiledevice3.services.os_trace import OsTraceService
from pymobiledevice3.services.syslog import SyslogService


@click.group()
def cli():
    """ apps cli """
    pass


@cli.group()
def syslog():
    """ syslog options """
    pass


@syslog.command('live-old', cls=Command)
def syslog_live_old(lockdown):
    """ view live syslog lines in raw bytes form from old relay """
    for line in SyslogService(lockdown=lockdown).watch():
        print(line)


@syslog.command('live', cls=Command)
@click.option('-o', '--out', type=click.File('wt'), help='log file')
@click.option('--nocolor', is_flag=True, help='disable colors')
@click.option('--pid', type=click.INT, default=-1, help='pid to filter. -1 for all')
@click.option('-m', '--match', help='match expression')
@click.option('-i', '--insensitive', is_flag=True, help='treat the match expression as case insensitive')
@click.option('include_label', '--label', is_flag=True, help='should include label')
def syslog_live(lockdown, out, nocolor, pid, match, insensitive, include_label):
    """ view live syslog lines """

    log_level_colors = {
        'Notice': 'white',
        'Error': 'red',
        'Fault': 'red',
        'Warning': 'yellow',
    }

    for syslog_entry in OsTraceService(lockdown=lockdown).syslog(pid=pid):
        syslog_pid = syslog_entry.pid
        timestamp = syslog_entry.timestamp
        level = syslog_entry.level
        filename = syslog_entry.filename
        image_name = os.path.basename(syslog_entry.image_name)
        message = syslog_entry.message
        process_name = os.path.basename(filename)
        label = ''

        if (pid != -1) and (syslog_pid != pid):
            continue

        if syslog_entry.label is not None:
            label = f'[{syslog_entry.label.bundle_id}][{syslog_entry.label.identifier}]'

        if not nocolor:
            timestamp = colored(str(timestamp), 'green')
            process_name = colored(process_name, 'magenta')
            if len(image_name) > 0:
                image_name = colored(image_name, 'magenta')
            syslog_pid = colored(syslog_entry['pid'], 'cyan')

            if level in syslog_entry:
                level = colored(level, log_level_colors[level])

            label = colored(label, 'cyan')
            message = colored(syslog_entry['message'], 'white')

        line_format = '{timestamp} {process_name}{{{image_name}}}[{pid}] <{level}>: {message}'

        if include_label:
            line_format += f' {label}'

        line = line_format.format(timestamp=timestamp, process_name=process_name, image_name=image_name, pid=syslog_pid,
                                  level=level, message=message)

        if match is not None:
            match_line = line
            if insensitive:
                match = match.lower()
                match_line = line.lower()
            if match not in match_line:
                continue
            else:
                if not nocolor:
                    match_line = match_line.replace(match, colored(match, attrs=['bold', 'underline']))
                    line = match_line

        print(line)

        if out:
            out.write(line)


@syslog.command('archive', cls=Command)
@click.argument('out', type=click.File('wb'))
def syslog_archive(lockdown, out):
    """
    create PAX archive.
    use `pax -r < filename` for extraction.
    """
    result, tar = OsTraceService(lockdown=lockdown).create_archive()
    out.write(tar)



#!/usr/bin/env python3
import logging
import plistlib
import struct
from datetime import datetime

from construct import Struct, Bytes, Int32ul, CString, Optional, Enum, Byte, Adapter
from pymobiledevice3.exceptions import PyMobileDevice3Exception
from pymobiledevice3.lockdown import LockdownClient

CHUNK_SIZE = 4096
TIME_FORMAT = '%H:%M:%S'
SYSLOG_LINE_SPLITTER = '\n\x00'


class TimestampAdapter(Adapter):
    def _decode(self, obj, context, path):
        return datetime.fromtimestamp(obj.seconds + (obj.microseconds / 1000000))

    def _encode(self, obj, context, path):
        return list(map(int, obj.split(".")))


timestamp_t = Struct(
    'seconds' / Int32ul,
    Bytes(4),
    'microseconds' / Int32ul
)

syslog_t = Struct(
    Bytes(9),
    'pid' / Int32ul,
    Bytes(42),
    'timestamp' / TimestampAdapter(timestamp_t),
    Bytes(1),
    'level' / Enum(Byte, Notice=0, Error=0x10, Fault=0x11),
    Bytes(1),
    Bytes(60),
    'filename' / CString('utf8'),
    'image_name' / CString('utf8'),
    'message' / CString('utf8'),
    'label' / Optional(Struct(
        'bundle_id' / CString('utf8'),
        'identifier' / CString('utf8')
    )),
)


class OsTraceService(object):
    SERVICE_NAME = 'com.apple.os_trace_relay'

    def __init__(self, lockdown: LockdownClient):
        self.logger = logging.getLogger(__name__)
        self.lockdown = lockdown
        self.c = self.lockdown.start_service(self.SERVICE_NAME)

    def get_pid_list(self):
        self.c.send_plist({'Request': 'PidList'})

        # ignore first received unknown byte
        self.c.recvall(1)

        response = self.c.recv_prefixed()
        return plistlib.loads(response)

    def create_archive(self) -> tuple:
        self.c.send_plist({'Request': 'CreateArchive'})

        # ignore first received unknown byte
        self.c.recvall(1)

        plist_response = plistlib.loads(self.c.recv_prefixed())

        # ignore first received unknown byte
        self.c.recvall(1)

        pax = self.c.recv_prefixed()

        return plist_response, pax

    def syslog(self, pid=-1):
        self.c.send_plist({'Request': 'StartActivity', 'MessageFilter': 65535, 'Pid': pid, 'StreamFlags': 60})

        length_length, = struct.unpack('<I', self.c.recvall(4))
        length = int(self.c.recvall(length_length)[::-1].hex(), 16)
        response = plistlib.loads(self.c.recvall(length))

        if response.get('Status') != 'RequestSuccessful':
            raise PyMobileDevice3Exception(f'got invalid response: {response}')

        while True:
            assert b'\x02' == self.c.recvall(1)
            length, = struct.unpack('<I', self.c.recvall(4))
            line = self.c.recvall(length)
            entry = syslog_t.parse(line)
            yield entry



from pymobiledevice3.services.dvt.tap import Tap


class ActivityTraceTap(Tap):
    IDENTIFIER = 'com.apple.instruments.server.services.activitytracetap'

    def __init__(self, dvt):
        # TODO: reverse [DTOSLogLoader _handleRecord:]

        config = {
            'bm': 0,  # buffer mode
            'combineDataScope': 0,
            'machTimebaseDenom': 3,
            'machTimebaseNumer': 125,
            'onlySignposts': 0,
            'pidToInjectCombineDYLIB': "-1",
            'predicate': "(messageType == info OR messageType == debug OR messageType == default OR "
                         "messageType == error OR messageType == fault)",
            'signpostsAndLogs': 1,
            'targetPID': "-3",
            'trackExpiredPIDs': 1,
            'ur': 500,
        }

        super().__init__(dvt, self.IDENTIFIER, config)
